(import
  (chicken io)
  (chicken process-context)
  (chicken file)
  (chicken pathname)
  (chicken string)
  (chicken pretty-print)
  srfi-1
  matchable
  mars parser)


;;; Games and Matches
;;

(define (file->prog file)
  (string->prog (with-input-from-file file read-string)))

(define (score-challenger-matches spec challenger-prog other-progs)
  (map 
   (lambda (other-prog)
     (score-match spec challenger-prog other-prog))
   other-progs))

(define (score-match spec prog1 prog2)
  (print "Matching " (prog-name prog1) " against " (prog-name prog2))
  (let ((tally
         (foldl
          (lambda (score-a score-b)
            (list (+ (car score-a) (car score-b))
                  (+ (cadr score-a) (cadr score-b))))
          (list 0 0)
          (let loop ((remaining (spec-games-per-match spec))
                     (results '()))
            (if (> remaining 0)
                (loop (- remaining 1)
                      (cons (score-game spec prog1 prog2)
                            results))
                results)))))
    (let ((prog1-name (prog-name prog1))
          (prog2-name (prog-name prog2)))
      `(((,prog1-name ,prog2-name) ,(car tally))
        ((,prog2-name ,prog1-name) ,(cadr tally))))))

(define (score-game spec prog1 prog2)
  (let* ((core (make-core (spec-core-size spec)))
         (queues (install-progs core (list prog1 prog2)))
         (result (run-mars core queues (spec-game-length spec))))
    (cond 
          ((null? result) (error "Invalid game result."))
          ((= (length result) 1)
           (let ((winner-name (caar result)))
             (if (eq? winner-name (prog-name prog1))
                 '(3 0)
                 '(0 3))))
          (else
           '(1 1)))))
    
(define (challenge hill-dir challenger-file)
  (let* ((spec (load-spec hill-dir))
         (scores (load-scores hill-dir))
         (challenger-prog (file->prog challenger-file))
         (hill-progs (map file->prog (hill-files hill-dir))))
    (if (memq (prog-name challenger-prog) (map prog-name hill-progs))
        (print "Challenger already on hill!")
        (let ((new-scores (append (score-challenger-matches spec challenger-prog hill-progs)
                                  scores)))
          (save-scores new-scores hill-dir)
          (hill-add-file challenger-file (prog-name challenger-prog) hill-dir)))))
  

;;; Hill initialization and specs
;;

(define (load-scores dir)
  (with-input-from-file (make-pathname dir "scores") read))

(define (save-scores scores dir)
  (with-output-to-file (make-pathname dir "scores")
    (lambda () (pretty-print scores))))

(define (load-spec dir)
  (with-input-from-file (make-pathname dir "spec") read))

(define (hill-files dir)
  (glob (make-pathname dir "*.red")))

(define (hill-add-file file name dir)
  (copy-file file (make-pathname dir (->string name) ".red")))

(define (make-spec core-size match-length games-per-match hill-size)
  (list 'spec hill-size core-size match-length games-per-match))

(define (spec? spec)
  (and (pair? spec) (eq? (car spec) 'spec)))

(define (spec-hill-size spec) (list-ref spec 1))
(define (spec-core-size spec) (list-ref spec 2))
(define (spec-game-length spec) (list-ref spec 3))
(define (spec-games-per-match spec) (list-ref spec 4))

(define (init-hill-dir dir hill-size core-size game-length games-per-match)
  (if (or (not (directory-exists? dir)) (not (file-writable? dir)))
      (print "Directory " dir " doesn't exist or is not writable.")
      (if (not (null? (glob (make-pathname dir "*"))))
          (print "Directory " dir " exists but is non-empty.")
          (begin
            (with-output-to-file (make-pathname dir "spec")
              (lambda ()
                (print ";; Hill specifications.")
                (print ";; ('spec hill-size core-size game-length games-per-match\n")
                (pp (make-spec core-size game-length games-per-match hill-size))))
            (with-output-to-file (make-pathname dir "scores")
              (lambda ()
                (pp '())))))))

;;;; Main ;;;;

;; Default values

(define default-core-size 8000)
(define default-game-length 80000)
(define default-games-per-match 1)
(define default-hill-size 10)

(define (print-usage)
  (let ((binary (pathname-file (car (argv)))))
    (print "King of the Hill Tournament Manager")
    (print "\nUsage:\t" binary " hill-directory challenger-file")
    (print "\t" binary " [-h|--help]")
    (print "\t" binary " [-i|--init] hill-directory [hill-size [core-size game-length games-per-match]]")
    (print "\nDefault values are as follows:\n"
           "\thill-size: " default-hill-size "\n"
           "\tcore-size: " default-core-size "\n"
           "\tgame-length: " default-game-length "\n"
           "\tgames-per-match: " default-games-per-match)))

(define (main)
  (match (cdr (argv))
    ((or () ((or "-h" "--help")))
     (print-usage))
    (((or "-i" "--init") dir hill-size core-size game-length games-per-match)
     (init-hill-dir dir hill-size core-size game-length games-per-match))
    (((or "-i" "--init") dir)
     (init-hill-dir dir
                    default-hill-size
                    default-core-size
                    default-game-length
                    default-games-per-match))
    (((or "-i" "--init") dir hill-size)
     (init-hill-dir dir
                    hill-size
                    default-core-size
                    default-game-length
                    default-games-per-match))
    ((hill-dir challenger-file)
     (challenge hill-dir challenger-file))
    (else
     (print "Invalid arguments: " (apply conc else)))))

;; (main)
